"""Dataset classes compliant to the Pytorch standard."""

# !/usr/bin/env python
# -*- coding: utf-8 -*-

# builtins
import logging
import pathlib
from datetime import date

# externals
import torch
import numpy as np
import xarray as xr
import dask.array as da

# locals
from climax.core.constants import (ERA5_VARIABLES, ERA5_PRESSURE_LEVELS,
                                   ERA5_P_VARIABLE_NAME, ERA5_S_VARIABLE_NAME,
                                   PROJECTION)
from pysegcnn.core.utils import search_files

# module level logger
LOGGER = logging.getLogger(__name__)


class EoDataset(torch.utils.data.Dataset):

    @staticmethod
    def to_tensor(x, dtype=torch.float32):
        """Convert ``x`` to :py:class:`torch.Tensor`.

        Parameters
        ----------
        x : array_like
            The input data.
        dtype : :py:class:`torch.dtype`
            The data type used to convert ``x``.

            The modified class labels.
        Returns
        -------
        x : `torch.Tensor`
            The input data tensor.

        """
        return torch.tensor(x, dtype=dtype)

    @staticmethod
    def encode_cyclical_features(feature, max_val):
        """Encode a cyclical feature to the range [-1, 1].

        Parameters
        ----------
        feature : :py:class:`numpy.ndarray`
            The cyclcical feature to encode.
        max_val : TYPE
            Maximum physically possible value of ``feature``.

        Returns
        -------
        encoded : `tuple` [:py:class:`numpy.ndarray`]
            The encoded feature in the range [-1, 1].

        """
        return (da.sin(2 * np.pi * feature / max_val).astype(np.float32),
                da.cos(2 * np.pi * feature / max_val).astype(np.float32))

    @staticmethod
    def add_coordinates(array, dims=('time', 'y', 'x')):
        return (dims, array)

    @staticmethod
    def repeat_along_axis(array, repeats, axis):
        return da.repeat(da.array(array), repeats, axis)

    @staticmethod
    def encode_doys(ds, **kwargs):

        # compute day of the year
        doys = ds.time.values.astype('datetime64[D]')
        doys = da.asarray(
            [date.timetuple(doy.astype(date)).tm_yday for doy in doys])

        # reshape doys to correct shape: from (t,) to (t, y, x)
        # this expands the doy values to each pixel (y, x)
        target = (len(doys), len(ds.y), len(ds.x))
        repeat = int(target[-1] * target[-2])

        # encode day of the year as cyclical feature: convert to dask array
        sin_doy, cos_doy = EoDataset.encode_cyclical_features(doys, 365)

        # lazily repeat encoded doys along time
        sin_doy, cos_doy = (
            EoDataset.repeat_along_axis(sin_doy, repeat, 0).reshape(target),
            EoDataset.repeat_along_axis(cos_doy, repeat, 0).reshape(target))

        return {'sin_doy': EoDataset.add_coordinates(sin_doy, **kwargs),
                'cos_doy': EoDataset.add_coordinates(cos_doy, **kwargs)}

    @staticmethod
    def state_file(model, predictand, predictors, plevels, dem=False,
                   doy=False):

        # naming convention:
        # <model>_<predictand>_<Ppredictors>_<plevels>_<Spredictors>.pt
        Ppredictors = ''.join([ERA5_P_VARIABLE_NAME[p] for p in predictors if p
                               in ERA5_P_VARIABLE_NAME])
        Spredictors = ''.join([ERA5_S_VARIABLE_NAME[p] for p in predictors if p
                               in ERA5_S_VARIABLE_NAME])
        plevels = [str(p) for p in plevels]
        state_file = '_'.join([model.__name__, str(predictand), Ppredictors,
                               *plevels, Spredictors])

        # check whether digital elevation model and day of year were used
        state_file = '_'.join([state_file, 'dem']) if dem else state_file
        state_file = '_'.join([state_file, 'doy']) if doy else state_file

        # add file extension: .pt
        return '.'.join([state_file, 'pt'])


class NetCDFDataset(EoDataset):

    def __init__(self, X, y=None, normalize=True, doy=False):

        # whether to include the day of the year as predictor variable
        if doy:
            # add doy to set of predictor variables
            LOGGER.info('Adding day of the year to predictor variables ...')
            X = X.assign(self.encode_doys(X))

        # NetCDF dataset containing predictor variables (ERA5)
        self.X = X.to_array().values.swapaxes(0, 1)

        # whether to normalize the training data to [0, 1] or standardize to
        # mean=0, std=1
        if normalize:
            # min-max-scaling: x in [0, 1]
            LOGGER.info('Normalizing data to [0, 1] ...')
            self.X -= self.X.min(axis=(0, -1, -2), keepdims=True)
            self.X /= self.X.max(axis=(0, -1, -2), keepdims=True)
        else:
            # standard scaling: mean=0, std=1
            LOGGER.info('Standardizing data to mean=0, std=1 ...')
            mean = np.nanmean(self.X, axis=(0, -1, -2), keepdims=True)
            std = np.nanstd(self.X, axis=(0, -1, -2), keepdims=True, ddof=1)
            self.X = (self.X - mean) / std

        # convert predictors to torch.Tensor
        self.X = self.to_tensor(self.X)

        # NetCDF dataset containing target variable (OBS)
        self.y = y
        if y is not None:
            self.y = self.to_tensor(y.to_array().values.swapaxes(0, 1))

    def __len__(self):
        return self.X.shape[0]

    def __getitem__(self, idx):
        # return a training sample given an index
        return (self.X[idx, ...],
                self.y[idx, ...] if self.y is not None else torch.tensor(0))


class ERA5Dataset(EoDataset):

    # compression for NetCDF
    comp = dict(dtype='float32', complevel=5, zlib=True)

    def __init__(self, root_dir, variables, plevels):

        # root directory: search for ERA5 files
        self.root_dir = pathlib.Path(root_dir)

        # valid variable names
        self.variables = [var for var in variables if var in ERA5_VARIABLES]

        # pressure levels
        self.plevels = [pl for pl in plevels if pl in ERA5_PRESSURE_LEVELS]

    def merge(self, **kwargs):
        # search dataset for each variable in root directory
        datasets = [search_files(self.root_dir.joinpath(var), '.nc$').pop() for
                    var in self.variables]

        # iterate over input datasets and select pressure levels
        LOGGER.info('Variables: {}'.format(', '.join(self.variables)))
        LOGGER.info('Pressure levels (hPa): {}.'.format(
            ', '.join([str(pl) for pl in self.plevels])))
        predictors = []
        for ds in datasets:
            # read dataset
            ds = xr.open_dataset(ds, **kwargs)

            # check if the dataset is defined on pressure levels or single
            # levels
            if 'level' in ds.dims:
                # iterate over pressure levels to use
                for pl in self.plevels:
                    # check if pressure level is available
                    if pl not in ds.level:
                        LOGGER.info('Pressure level {:d} hPa not available.'
                                    .format(pl))
                        continue

                    # select pressure level and drop unnecessary dimensions
                    level = ds.sel(level=pl).drop('level')

                    # rename variable including corresponding pressure level
                    level = level.rename({k: '_'.join([k, str(pl)]) for k in
                                          level.data_vars if k != PROJECTION})

                    # append current pressure level to list of predictors
                    predictors.append(level)
            else:
                # append single level dataset to list of predictors
                predictors.append(ds)

        # drop projection variable, if present
        predictors = [ds.drop_vars(PROJECTION) for ds in predictors if
                      PROJECTION in ds.data_vars]

        # merge final predictor dataset
        return xr.merge(predictors)
