"""Dynamical climate downscaling using deep convolutional neural networks."""

# !/usr/bin/env python
# -*- coding: utf-8 -*-

# builtins
import time
import logging
from datetime import timedelta
from logging.config import dictConfig

# externals
import torch
import numpy as np
import xarray as xr
from sklearn.model_selection import train_test_split
from torch.utils.data import DataLoader

# locals
from pysegcnn.core.utils import search_files, img2np
from pysegcnn.core.trainer import NetworkTrainer, LogConfig
from pysegcnn.core.models import Network
from pysegcnn.core.logging import log_conf
from climax.core.dataset import ERA5Dataset, NetCDFDataset
from climax.main.config import (ERA5_PLEVELS, ERA5_PREDICTORS, PREDICTAND,
                                CALIB_PERIOD, DOY, SHUFFLE, BATCH_SIZE, LR,
                                LAMBDA, NORM, TRAIN_CONFIG, NET, LOSS, FILTERS,
                                OVERWRITE, DEM)
from climax.main.io import ERA5_PATH, OBS_PATH, DEM_PATH, MODEL_PATH

# module level logger
LOGGER = logging.getLogger(__name__)


if __name__ == '__main__':

    # initialize timing
    start_time = time.monotonic()

    # initialize network filename
    state_file = ERA5Dataset.state_file(
        NET, PREDICTAND, ERA5_PREDICTORS, ERA5_PLEVELS, dem=DEM, doy=DOY)
    state_file = MODEL_PATH.joinpath(state_file)

    # initialize logging
    log_file = MODEL_PATH.joinpath(state_file.name.replace('.pt', '_log.txt'))
    if log_file.exists():
        log_file.unlink()
    dictConfig(log_conf(log_file))

    # initialize downscaling
    LogConfig.init_log('Initializing downscaling for period: {}'.format(
        ' - '.join([str(CALIB_PERIOD[0]), str(CALIB_PERIOD[-1])])))

    # check if model exists
    if state_file.exists() and not OVERWRITE:
        # load pretrained network
        net, _ = Network.load_pretrained_model(state_file, NET)
    else:
        # initialize ERA5 predictor dataset
        LogConfig.init_log('Initializing ERA5 predictors.')
        Era5 = ERA5Dataset(ERA5_PATH.joinpath('ERA5'), ERA5_PREDICTORS,
                           plevels=ERA5_PLEVELS)
        Era5_ds = Era5.merge()

        # initialize OBS predictand dataset
        LogConfig.init_log('Initializing observations for predictand: {}'
                           .format(PREDICTAND))

        # check whether to joinlty train tasmin and tasmax
        if PREDICTAND == 'tas':
            # read both tasmax and tasmin
            tasmax = xr.open_dataset(
                search_files(OBS_PATH.joinpath('tasmax'), '.nc$').pop())
            tasmin = xr.open_dataset(
                search_files(OBS_PATH.joinpath('tasmin'), '.nc$').pop())
            Obs_ds = xr.merge([tasmax, tasmin])
        else:
            # read in-situ gridded observations
            Obs_ds = search_files(OBS_PATH.joinpath(PREDICTAND), '.nc$').pop()
            Obs_ds = xr.open_dataset(Obs_ds)

        # whether to use digital elevation model
        if DEM:
            # digital elevation model: Copernicus EU-Dem v1.1
            LogConfig.init_log('Initializing digital elevation model.')
            dem = search_files(DEM_PATH, '^eu_dem_v11_stt.nc$').pop()
            dem = np.expand_dims(img2np(dem), axis=0)
            dem = ERA5Dataset.repeat_along_axis(dem, len(Era5_ds.time), 0)
            dem = ERA5Dataset.add_coordinates(dem)
            dem = xr.Dataset(data_vars={'elevation': dem},
                             coords=Era5_ds.coords)

            # add dem to set of predictor variables
            Era5_ds = xr.merge([Era5_ds, dem])

        # initialize training data
        LogConfig.init_log('Initializing training data.')

        # split calibration period into training and validation period
        train, valid = train_test_split(CALIB_PERIOD, shuffle=False)

        # training and validation dataset
        Era5_train, Obs_train = Era5_ds.sel(time=train), Obs_ds.sel(time=train)
        Era5_valid, Obs_valid = Era5_ds.sel(time=valid), Obs_ds.sel(time=valid)

        # create PyTorch compliant dataset and dataloader instances for model
        # training
        train_ds = NetCDFDataset(Era5_train, Obs_train, normalize=NORM,
                                 doy=DOY)
        valid_ds = NetCDFDataset(Era5_valid, Obs_valid, normalize=NORM,
                                 doy=DOY)
        train_dl = DataLoader(train_ds, batch_size=BATCH_SIZE, shuffle=SHUFFLE,
                              drop_last=False)
        valid_dl = DataLoader(valid_ds, batch_size=BATCH_SIZE, shuffle=SHUFFLE,
                              drop_last=False)

        # instanciate network from scratch
        outputs = 3 if PREDICTAND == 'pr' else len(Obs_ds.data_vars)
        net = NET(state_file, train_ds.X.shape[1], outputs, filters=FILTERS)

        	# initialize optimizer
        optimizer = torch.optim.Adam(net.parameters(), lr=LR,
                                     weight_decay=LAMBDA)

        # initialize network trainer
        trainer = NetworkTrainer(net, optimizer, net.state_file, train_dl,
                                 valid_dl, loss_function=LOSS, **TRAIN_CONFIG)

        # train model
        state = trainer.train()


    # log execution time of script
    LogConfig.init_log('Execution time of script {}: {}'
                       .format(__file__, timedelta(seconds=time.monotonic() -
                                                   start_time)))
